#include <unistd.h>

extern int is_terminal;

void store_terminal_modes();
int reset_terminal_modes();
void disable_terminal_signals();
void enable_terminal_signals();
int change_terminal_owner(uid_t uid, int is_login,
          const char **callname, const char **err_descr);
void restore_terminal_owner();
int make_process_unkillable(const char **callname
        , const char **err_descr);
void make_process_killable();
void usage();
void parse_command_line(int argc, char *argv[]
        , int *is_login, const char **user, const char **command);

